<?php defined('KERNEL_INCLUDED') or die('kernel should be included first');
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder, Arthur Charguraud

//-----------------------------------------------------------------
/// @file
/// @brief Data structures and constants used in DENIF
//-----------------------------------------------------------------


//-----------------------------------------------------------------
/// Information about a SQL table
//-----------------------------------------------------------------

class ObjectTableInfo
{
	public $sName;
	public $aFullText;
	
	
	private function __construct($sName, $sFullTexts)
	{
		$this->sName     = $sName;
		$this->aFullText = getList($sFullTexts);
	}
	
	private static $a = null;
	
	public static function get($iTable)
	{
		if (self::$a === null)
			self::$a = array(
				tableNode   => new ObjectTableInfo('node',   'sTitle;sDetails;sURL'),
				tableCursus => new ObjectTableInfo('cursus', 'sShortName;sFullName;sURL'),
				tableTheme  => new ObjectTableInfo('theme',  'sTitle'),
				tableHuman  => new ObjectTableInfo('human',  'sName;sURL'),
			);
		
		return self::$a[$iTable];
	}
}


//-----------------------------------------------------------------
/// Information about an object type.
//-----------------------------------------------------------------

class ObjectTypeInfo
{
	public $table, $sOrderBy, $sName;
	public $sAlone, $sPlural, $sUnUne, $sLeLa, $sDuDe;
	
	/// @brief List of the properties that can be linked to the object type.
	///
	/// Ignored for properties. For nodes, it is an associative array
	/// <tt>(iProperty => nbAdd)</tt> where \c iProperty is a property type
	/// and \c nbAdd is an integer having the following meaning:
	/// - If <tt>= 0</tt>, the node type can't have links to this property
	///   (same as if the array does not contain an entry for \c iProperty).
	/// - If <tt>\> 0</tt>, specifies the number of links adding combo-boxes to display
	//    in the node editing page.
	/// - If <tt>\< 0</tt>, the node editing page uses a text area to specify
	///   the links list. Each line of the text area contains the name of a link.
	/// If <tt>nbAdd != 0</tt>, a link allowing the user to create a new property
	/// and link to to the edited node is added to the node editing page.
	public $aProperty;
	
	
	private function __construct($iTable, $sOrderBy, $sName, $sAlone, $sUnUne, $sLeLa, $sDuDe,
		array $aProperty = array())
	{
		$this->table     = ObjectTableInfo::get($iTable);
		$this->sOrderBy  = $sOrderBy;
		$this->sName     = $sName;
		$this->sAlone    = $sAlone;
		$this->sPlural   = $sAlone . ((substr($sAlone, -1) == 's') ? '' : 's');
		$this->sUnUne    = $sUnUne . ' ' . $sAlone;
		$this->sLeLa     = $sLeLa . $sAlone;
		$this->sDuDe     = $sDuDe . $sAlone;
		$this->aProperty = $aProperty;
	}
	
	public function instantiate()
	{
		return new $this->table->sName;
	}
	
	
	private static $a;
	
	public static function get($iObject)
	{
		if (self::$a === null)
			self::$a = array(
				nodeResource   => new ObjectTypeInfo(tableNode,   'sTitle',     'Resource', 'ressource',  'une', 'la ', 'de la ',
					array(propertyTheme => 2, propertyCursus => 2)),
				nodeSession    => new ObjectTypeInfo(tableNode,   'iOrder',     'Session',  'session',    'une', 'la ', 'de la ',
					array()),
				nodeDocGroup   => new ObjectTypeInfo(tableNode,   'iOrder',     'DocGroup', 'groupe',     'un',  'le ', 'du ',
					array(propertyHuman => -1)),
				nodeDocument   => new ObjectTypeInfo(tableNode,   'iOrder',     'Document', 'document',   'un',  'le ', 'du ',
					array(propertyHuman => -1)),
				propertyCursus => new ObjectTypeInfo(tableCursus, 'sShortName', 'Cursus',   'cursus',     'un',  'le ', 'du '),
				propertyTheme  => new ObjectTypeInfo(tableTheme,  'sTitle',     'Theme',    'thme',      'un',  'le ', 'du '),
				propertyHuman  => new ObjectTypeInfo(tableHuman,  'sName',      'Human',    'enseignant', 'un',  'l',  'de l'),
			);
		
		return defaultElement(self::$a, $iObject, null);
	}
}


// Define some enums:
// - nodeXXX
// - propertyXXX
// - humanXXX
// - pdlXXX (PingDownLoad)
/// \c nodeXXX and \c propertyXXX do not share the same values,
/// allowing the type of any object to be refered by an integer value.
defineEnum('node',      'Resource;Session;DocGroup;Document', true, 1);
defineEnum('property',  'Cursus;Theme;Human', true, 100);
defineEnum('table',     'Node;Cursus;Theme;Human', true);
defineEnum('human',     'Unknown;Course;TD;Author', true);
defineEnum('nodeFlag',  'Visible;Final;CacheRequested;CacheVisible;CacheAuthorized', true);
defineEnum('pdl',       'Ignore;Ping;Download', true);


//-----------------------------------------------------------------
/// Information about a node flag
//-----------------------------------------------------------------

class NodeFlag
{
	const sImages = 'inherit;yes;no;unknown';
	
	public $sShortName;     ///< The name of the flag, with a lowercase first letter.
	public $sPHPName;       ///< The name of the flag, with a type prefix (PHP member name).
	public $sImagePrefix;   ///< Prefix of the flag images name.
	public $aValue;         ///< User-friendly name of the possible values.
	                        /// <tt>%s</tt> in the first value will be replaced by the name of the inherited value.
	public $sName;          ///< User-friendly name of the flag.
	public $sDescription;   ///< User-friendly description of the flag.
	
	
	private function __construct($sShortName, $sPHPName, $sImagePrefix, $sChoices, $sValues, $sName, $sDescription)
	{
		defineEnum($sShortName, $sChoices, true);
		
		$this->sShortName   = $sShortName;
		$this->sPHPName     = $sPHPName;
		$this->sImagePrefix = $sImagePrefix;
		$this->aValue       = getList($sValues);
		$this->sName        = $sName;
		$this->sDescription = $sDescription;
	}
	
	
	private static $a;
	
	public static function getAll()
	{
		if (self::$a === null)
			self::$a = array(
				new NodeFlag('visible',   'iVisible',         'visible',          'Inherit;Yes;No',         'Hriter (%s);Visible;Masqu',
					'Visible', 'Lobjet est visible aux utilisateurs&nbsp;?'),
				new NodeFlag('final',     'iFinal',           'final',            'Inherit;Yes;No;Temp',    'Hriter (%s);Final;Propos',
					'Final', 'Lobjet a t valid par un administrateur&nbsp;?'),
				new NodeFlag('cacheReq',  'iCacheRequested',  'cache-requested',  'Inherit;Yes;No',         'Hriter (%s);Cache demand;Cache non demand',
					'Cache demand', 'Tlcharger le fichier ds que possible&nbsp;?'),
				new NodeFlag('cacheVis',  'iCacheVisible',    'cache-visible',    'Inherit;Yes;No',         'Hriter (%s);Cache visible;Cache invisible',
					'Cache visible', 'Les utilisateurs peuvent tlcharger la version en cache&nbsp;?'),
				new NodeFlag('cacheAuth', 'iCacheAuthorized', 'cache-authorized', 'Inherit;Yes;No;Unknown', 'Hriter (%s);Cache autoris;Cache interdit;Statut inconnu',
					'Cache autoris', 'Lauteur a autoris la mise en cache&nbsp;?'),
			);
		
		return self::$a;
	}
};

NodeFlag::getAll();


/// Concatenate path of objects given as arguments
function concatPathPart(array $aNode)
{
	$aPathPart = projectTo($aNode, 'sPathPart');
	return (in_array('', $aPathPart)) ? '' : path::concat($aPathPart);
}


//-----------------------------------------------------------------
/// Base class for nodes and properties
//-----------------------------------------------------------------

abstract class ObjectBase
{
	/// ID of the object, unique among all objects belonging to the same SQL table.
	public $id;
	
	/// @brief Name of the HTML achor the object can use.
	/// This is the object ID, prefixed by an underscore
	/// because IDs beginning by a digit are XHTML-invalid.
	public function getAnchor() { return '_' . $this->id; }
	
	
	//-----------------------------------------------------------------
	/// @name Instance-independant information (would be virtual static functions)
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Type of the object (node* or property* constant).
	public abstract function getType();
	
	/// @brief #ObjectTypeInfo object of the node or property.
	public function getInfo() { return ObjectTypeInfo::get($this->getType()); }
	
	/// @brief Suffix of the PHP page containing the editing form.
	/// For example, <tt>'foo'</tt> points to <tt>'edit_foo.php'</tt>.
	public abstract function getFormName();
	
	
	//-----------------------------------------------------------------
	/// @name Linking
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief HTML link pointing to a page displaying the object.
	/// If the object has no URL, HTML text is returned.
	public abstract function getLinkTitle();
	
	/// #target object pointing to a page displaying the object.
	public abstract function getLinkTarget();
	
	/// @brief Return the absolute URL of the object,
	/// an empty string if no absolute URL is available.
	public function getAbsoluteURL()
	{
		if (!isset($this->sURL))
			return '';
		
		return (@parse_url($this->sURL, PHP_URL_SCHEME))
			? $this->sURL : '';
	}
	
	
	/// @brief CSS class name used by external links.
	/// House for teachers home page, basic icon for all other links.
	public function getLinkOutClassName() { return 'external'; }
	
	/// @brief Tooltip text used by external links.
	/// Teachers have an home page, other objects have a Web page.
	public function getLinkOutToolTip() { return 'Page Web'; }
	
	
	/// @brief Return an HTML link pointing to a page displaying the object.
	/// If \p $sTitle is \c null, the object title is used as link caption.
	public abstract function link(PageDenif $page, $sTitle = null, $sClassName = '');
	
	/// @brief External mode: use an icon as title.
	/// Internal mode: an empty string if the URL is empty.
	public function linkOutTo(PageDenif $page, $sURL, $bExternal = false, $sToolTip = '')
	{
		if ($bExternal)
			return ($sURL)
				? ' ' . $page->linkOut($sURL, html::nbsp(), 'action ' . $this->getLinkOutClassName(),
					($sToolTip) ? $sToolTip : $this->getLinkOutToolTip())
				: '';
		else{
			$sTitle = html::text($this->getLinkTitle());
			if ($sURL)
				return $page->linkOut($sURL, $sTitle, '', $sToolTip);
			else if ($sToolTip)
				return html::span($sTitle, attr::title($sToolTip));
			else
				return $sTitle;
		}
	}
	
	/// Reserved to classes having a <tt>$sURL</tt> member.
	public function linkOut(PageDenif $page, $bExternal = false, $sToolTip = '')
	{
		return $this->linkOutTo($page, $this->getAbsoluteURL(),
			$bExternal, $sToolTip);
	}
	
	
	/// Return a link to the object with an external link, if available.
	public function linkFull(PageDenif $page, $sClassName = '')
	{
		$s = $this->link($page, null, $sClassName);
		$sURL = $this->getAbsoluteURL();
		if ($sURL)
			$s .= $this->linkOutTo($page, $sURL, true);
		return $s;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Deleting
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Delete the object and the links concerning it.
	/// The database is leaved in a consistent state.
	public abstract function delete();
	
	/// @brief Return a boolean value indicating if the object can be safely deleted
	/// from database integrity point of view. No rights check is performed.
	public abstract function canDelete();
	
	/// @brief Return an HTML summary of the node. Used by delete confirmation page.
	public abstract function dumpBeforeDelete();
	
	/// Helper used by #dumpBeforeDelete() implementors.
	protected static function dumpField($sName, $sValue)
	{
		return html::div(html::span(ucfirst($sName) . '&nbsp;:', attr::className('field-label')) . ' ' . html::text($sValue));
	}
	
	/// @}
	
	
	/// @brief Check the sanity of the object taken alone
	/// (no database check, no links check).
	/// Return an array of problems description.
	public abstract function sanityCheck();
}


//-----------------------------------------------------------------
/// Class for all nodes: resources, sessions, doc. groups, documents
//-----------------------------------------------------------------

class node extends ObjectBase
{
	//-----------------------------------------------------------------
	/// @name Fields of the SQL table
	/// @{
	//-----------------------------------------------------------------
	
	public $iNode;         ///< @brief Type of the node, see node* constants.
	public $idParent;      ///< @brief ID of the parent node, should be 0 for root nodes (resources).
	public $iOrder;        ///< @brief Position of the node among its brothers, first is 0. Ignored for resources, which are alphabetically sorted.
	public $sTitle;        ///< @brief Title of the node, displayed everywhere.
	public $sDetails;      ///< @brief Internal details about the resource, for admins only.
	public $sPathPart;     ///< @brief File name for leaves (documents), directory for other nodes.
	                       /// Should not contain any character of #sDataPathPartInvalidCharsSpec.
	public $sURL;          ///< @brief File URL for documents, URL describing the node for other nodes.
	public $dtCreation;    ///< @brief Creation date of the node (unrelated to cache file).
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Node flags
	/// @{
	//-----------------------------------------------------------------
	
	public $iVisible;           ///< @brief Node is visible by guests?
	public $iFinal;             ///< @brief Node is finalized, or still editable by guests?
	public $iCacheRequested;    ///< @brief Documents should be downloaded to cache files?
	public $iCacheVisible;      ///< @brief Cache files are downloadable by guests?
	public $iCacheAuthorized;   ///< @brief Caching has been authorized by document owner?
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Cache-related data
	/// @{
	//-----------------------------------------------------------------
	
	public $dtPing;        ///< @brief Last ping date, valid even if last ping was negative, \c 0 if no ping.
	public $bPingOK;       ///< @brief Result of the last ping, \c false if unknown.
	public $iPingDownload; ///< @brief Indicate if the file can/should be ignored, pinged, or cached.
	
	public $bCached;       ///< @brief Indicate if the file is currently cached.
	                       /// This value is not used by the site and is not automatically synchronized.
	                       /// It can be manually updated by using the 'Check cached files' admin command.
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Data from other objects
	/// @{
	//-----------------------------------------------------------------
	
	public $parent;        ///< @brief Object of the parent node, #getDefault() for root nodes.
	public $aFlagInherit;  ///< @brief Inherited node flags, indexed by nodeFlags* constants.
	                       /// Contains no 'inherit' value but the inherited, effective value itself.
	
	/// @}
	
	
	public static $table;  ///< @brief #ObjectTableInfo object common to all node types.
	
	
	//-----------------------------------------------------------------
	/// @name Accessors
	/// @{
	//-----------------------------------------------------------------
	
	public function getType() { return $this->iNode; }
	public function getFormName() { return 'node'; }
	
	public function getLinkTitle() { return defaultValue($this->sTitle, '[sans titre]'); }
	public function getLinkTarget() { return new target('node.php', array('id' => $this->id)); }
	
	/// Title of the node displaying page.
	public function getPageTitle()
	{
		return ucfirst($this->getInfo()->sAlone) . '&nbsp;: ' . html::text($this->getLinkTitle());
	}
	
	/// @}
	
	
	public function sanityCheck()
	{
		$a = array();
		if (!$this->sTitle)  $a[] = 'titre vide';
		return $a;
	}
	
	
	/// @brief Return a formatted version of #$sDetails,
	/// in which URLs are replaced by links.
	/// The html::text() function should not be called on the returned string.
	public function formatDetails()
	{
		$s = regexp::replace('|https?://\S*[^\s,;:!\\?\\.]|',
			'<$0>', html::text($this->sDetails));
		return regexp::replace('|<(.*)>|U', html::link('$1', '$1'), $s);
	}
	
	
	/// @brief Return a formatted version of #$sTitle,
	/// used as filename in generated Zip archived.
	public function formatZipTitle()
	{
		$s = regexp::replace(regexp::encloseChars(path::sPartInvalidCharsSpec), '', $this->sTitle);
		
		$sFinal = '';
		foreach (regexp::split(regexp::enclose('\\s+'), $s) as $sPart) {
			if ($sFinal)
				$sFinal .= ' ';
			$sFinal .= $sPart;
			if (strlen($sFinal) >= maxLengthZipFileName)
				break;
		}
		
		$sFinal = trim($sFinal, ',;:.');
		return $sFinal . path::getExtensionDotted($this->sPathPart);
	}
	
	
	/// @brief Indicate if the cache file can be downloaded by a guest user,
	/// according to node flags.
	public function canGuestAccessCache()
	{
		return
			$this->aFlagInherit[nodeFlagCacheVisible]    == cacheVisYes &&
			$this->aFlagInherit[nodeFlagCacheAuthorized] != cacheAuthNo;
	}
	
	
	/// Condition to get all brothers of the node.
	public function getBrotherSqlCondition()
	{
		return ($this->iNode == nodeFirst)
			? array('iNode' => nodeFirst)
			: array('idParent' => $this->idParent);
	}
	
	
	//-----------------------------------------------------------------
	/// @name Data retrieving
	/// @{
	//-----------------------------------------------------------------
	
	public function getPropertiesLink($iProperty, $sMore = '', array $aTableMore = array())
	{
		return getPropertiesOf($iProperty,
			'AND ' . sql::formatField('link.idNode', $this->id) . ' ' . $sMore,
			array('link.*'), $aTableMore);
	}
	
	public function getPropertiesAndLink($iProperty, $sMore = '', array $aTableMore = array())
	{
		return getPropertiesAndLinkOf($iProperty,
			'AND ' . sql::formatField('link.idNode', $this->id) . ' ' . $sMore,
			$aTableMore);
	}
	
	public function getProperties($iProperty, $sMore = '', array $aTableMore = array())
	{
		return getPropertiesOf($iProperty,
			' AND ' . sql::formatField('link.idNode', $this->id) . ' ' . $sMore,
			$aTableMore);
	}
	
	
	/// @brief Load the children of given rank of the node.
	/// By default, the direct children are loaded.
	///
	/// The #$parent property of the children is set.
	///
	/// @param $iNodeLeaf the node type of the children to load,
	/// strictly negative to load the direct children.
	public function loadChildren($iNodeLeaf = -1)
	{
		if ($iNodeLeaf < 0)
			$iNodeLeaf = $this->iNode + 1;
		if ($iNodeLeaf > nodeLast)
			return array();
		
		$sTableLeaf = getNodeTable($iNodeLeaf);
		$aTable = array();
		$aWhere = array(getNodeTable($this->iNode) . '.id' => $this->id);
		sqlNodesChildren($aTable, $aWhere, $this->iNode, $iNodeLeaf);
		
		$aChild = sql::selectFetch($aTable,
			array($sTableLeaf . '.*'), sql::where($aWhere) .
			' ORDER BY ' . $sTableLeaf . '.' . ObjectTypeInfo::get($iNodeLeaf)->sOrderBy,
			self::$table->sName);
		
		foreach ($aChild as $nodeChild)
			$nodeChild->parent = $this;
		
		return $aChild;
	}
	
	/// @brief Load and return the parent of the node,
	/// \c null if then node is root. Store a link to the parent in the #$parent property.
	public function loadParent()
	{
		$this->parent = ($this->iNode == nodeResource)
			? self::getDefault() : getNode($this->idParent);
		return $this->parent;
	}
	
	/// @brief Return an array containing all ancestors of the node,
	/// indexed by their node type.
	/// The first element of the array is the root ancestor.
	/// The array is empty if the node is a resource.
	/// Links to parents are stored in the #$parent property.
	public function loadAllParents($bComputeInheritedFlags)
	{
		$aParent = array();
		$objParent = $this;
		foreach (getDefineRange('node') as $i) {
			if ($objParent->iNode == nodeFirst)
				break;
			$objParent = $objParent->loadParent();
			if (!$objParent)
				break;
			$aParent[$objParent->iNode] = $objParent;
		}
		$aParent = array_reverse($aParent, true);
		
		if ($bComputeInheritedFlags) {
			$nodeParent = node::getDefault();
			foreach ($aParent as $node) {
				$node->computeInheritedFlags($nodeParent);
				$nodeParent = $node;
			}
			$this->computeInheritedFlags($nodeParent);
		}
		
		return $aParent;
	}
	
	/// @brief Load all parents and children of the node, recursively.
	/// The children array is stored in a \c $aChild property.
	/// The inherited flags are computed.
	public function loadParentsChildrenRecurse($bLoadParents = true)
	{
		if ($bLoadParents)
			$this->loadAllParents(true);
		
		$this->aChild = $this->loadChildren();
		foreach ($this->aChild as $nodeChild) {
			$nodeChild->computeInheritedFlags($this);
			$nodeChild->loadParentsChildrenRecurse(false);
		}
	}
	
	
	/// @brief Get all property links about a resource and/or its child nodes.
	/// @return an associative array <tt>'link id' => 'link'</tt>.
	public function getLinksRecurse($iNodeFirst = -1, $iNodeLast = nodeLast)
	{
		$aLink = array();
		
		if ($iNodeFirst < 0)
			$iNodeFirst = $this->iNode;
		for ($iNodeLeaf = $iNodeFirst; $iNodeLeaf <= $iNodeLast; $iNodeLeaf++) {
			$aTable = array('link');
			$aWhere = array(
				'!link.idNode'                    => getNodeTable($iNodeLeaf) . '.id',
				getNodeTable($iNodeFirst) . '.id' => $this->id,
			);
			sqlNodesChildren($aTable, $aWhere, $iNodeFirst, $iNodeLeaf);
			
			$aLink += sql::selectFetchBy($aTable, 'id',
				array('link.*'), sql::where($aWhere));
		}
		
		return $aLink;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Linking
	/// @{
	//-----------------------------------------------------------------
	
	public function link(PageDenif $page, $sTitle = null, $sClassName = '')
	{
		return $page->linkTarget($this->getLinkTarget(),
			($sTitle === null) ? html::text($this->getLinkTitle()) : $sTitle,
			$sClassName);
	}
	
	/// @brief Return a icon link to the node page.
	public function linkDetails(PageDenif $page)
	{
		$sToolTip = 'Dtails';
		if ($page->isAdmin())
			$sToolTip .= ' (ID = ' . $this->id . ')';
		return DenifLink::createAction($this->getLinkTarget(), 'details', $sToolTip);
	}
	
	/// @brief Return a icon link to the Zip generation page.
	public function linkZip(PageDenif $page)
	{
		return DenifLink::createAction(new target('zip.php', array('idNode' => $this->id)), 'zip', 'Archive Zip des documents en cache');
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Deleting
	/// @{
	//-----------------------------------------------------------------
	
	public function delete()
	{
		self::deleteRecurse($this->iNode, $this->id, $this->sPathFull);
		sql::update(self::$table->sName, array('!iOrder' => 'iOrder-1'),
			sql::where($this->getBrotherSqlCondition()) . ' AND iOrder>' . $this->iOrder);
		return null;
	}
	
	private static function deleteRecurse($iNode, $id, $sPathFull)
	{
		$sTable = self::$table->sName;
		sql::delete($sTable, sql::whereId($id));
		sql::delete('link', sql::whereOne('idNode', $id));
		
		if ($iNode < nodeLast) {
			// Directory
			
			$iNode++;
			foreach (sql::selectFieldBy($sTable, 'id', 'sPathPart',
			         sql::whereOne('idParent', $id)) as $idChild => $sPathPart)
				self::deleteRecurse($iNode, $idChild, path::concat($sPathFull, $sPathPart));
			
			@rmdir($sPathFull);
			
		}else{
			// File
			
			@unlink($sPathFull);
		}
	}
	
	
	public function canDelete() { return true; }
	
	
	public function dumpBeforeDelete()
	{
		$info = $this->getInfo();
		$s = '';
		
		$s .= self::dumpField($info->sAlone, $this->sTitle);
		if ($this->sDetails)
			$s .= html::par($this->formatDetails());
		
		$iNodeRoot = $this->iNode;
		for ($iNodeLeaf = $iNodeRoot + 1; $iNodeLeaf <= nodeLast; $iNodeLeaf++) {
			$aTable = array();
			$aWhere = array(getNodeTable($iNodeRoot) . '.id' => $this->id);
			sqlNodesChildren($aTable, $aWhere, $iNodeRoot, $iNodeLeaf);
			
			$aNode = sql::selectFetch($aTable, array(getNodeTable($iNodeLeaf) . '.sTitle'), sql::where($aWhere));
			if ($aNode)
				$s .= self::dumpField(ObjectTypeInfo::get($iNodeLeaf)->sAlone . plural($aNode) . '  supprimer',
					concat(projectTo($aNode, 'sTitle'), ', '));
		}
		
		return $s;
	}
	
	/// @}
	
	
	/// @brief Return HTML-formatted description of the node.
	public function dump()
	{
		$aAncestor = $this->loadAllParents(true);
		$aAncestor[] = $this;
		
		$a = array();
		foreach ($aAncestor as $node) {
			$sNode = $node->getLinkTitle();
			if ($node->id == $this->id)
				$sNode = html::b($sNode);
			$a[] = sprintf('%s (%s, ID = %d)', $sNode, $node->getInfo()->sAlone, $node->id);
		}
		return concat($a, ' &rarr; ');
	}
	
	
	//-----------------------------------------------------------------
	/// @name Node flags and inheritance
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Compute the effective flags of the node.
	///
	/// When necessary, inherit the flags from the given parent.
	public function computeInheritedFlags(node $parent)
	{
		foreach (NodeFlag::getAll() as $iFlag => $nodeFlag) {
			$sPHPName = $nodeFlag->sPHPName;
			$iValue = $this->$sPHPName;
			$this->aFlagInherit[$iFlag] = ($iValue == 0) ? $parent->aFlagInherit[$iFlag] : $iValue;
		}
		
		$this->sPathFull = ($this->sPathPart)
			? path::concat($parent->sPathFull, $this->sPathPart)
			: $parent->sPathFull;
		
		$this->parent = $parent;
	}
	
	
	/// @brief Get a copy of the NodeFlag::getAll() array
	/// using the node values as default inherited values.
	/// The returned NodeFlag objects will have a \c $aImage property
	/// containing the CSS class name of all possible values.
	/// Each CSS class name specifies the image of the flag value.
	public function getNodeFlags()
	{
		$a = array();
		$aFlagImage = getList(NodeFlag::sImages);
		
		foreach (NodeFlag::getAll() as $iFlag => $nodeFlag) {
			$nodeFlag = clone $nodeFlag;
			$iValue = $this->aFlagInherit[$iFlag];
			
			$aValue = &$nodeFlag->aValue;
			$aValue[0] = sprintf($aValue[0], $aValue[$iValue]);
			
			$aImage = array();
			foreach ($aFlagImage as $sImage)
				$aImage[] = $nodeFlag->sImagePrefix . ' ' . $sImage;
			$aImage[0] .= '-' . $aFlagImage[$iValue];
			$nodeFlag->aImage = $aImage;
			
			$a[$iFlag] = $nodeFlag;
			
			unset($aValue);
		}
		
		return $a;
	}
	
	
	/// Return a \c node object having default values for inheritable flags.
	public static function getDefault()
	{
		$nodeDefault = new node();
		
		$nodeDefault->aFlagInherit = array();
		foreach (NodeFlag::getAll() as $iFlag => $nodeFlag)
			$nodeDefault->aFlagInherit[$iFlag] = constant($nodeFlag->sShortName . 'Default');
		
		$nodeDefault->sPathFull = sDataPath;
		
		return $nodeDefault;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Path part
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Apply the change of the #$sPathPart field to the file system:
	/// create/rename/delete the file or folder attached to the node.
	/// Return \c null if no error occured, an HTML error string otherwise.
	///
	/// @param $sPathPartNew new path part to give to the node.
	/// @param $aParent list of all parents of the node, excluding the node itself.
	/// Used to compute the full path of the node.
	/// @param $aBrotherPathPart path part of all brothers of the node, excluding the node itself.
	/// Used to prevent two brothers from having the same path part.
	public function onPathPartChanged($sPathPartNew, array $aParent, array $aBrotherPathPart)
	{
		$sPathPartOld = $this->sPathPart;
		if ($sPathPartOld == $sPathPartNew)
			return null;
		
		$bIsFile = ($this->iNode == nodeLast);
		$sFileOrDir = ($bIsFile) ? 'fichier' : 'dossier';
		
		if (!self::checkPathPart($sPathPartNew))
			return ucfirst($sFileOrDir) . '&nbsp;: portion de chemin daccs invalide&nbsp;: ' . html::kbdText($sPathPartNew);
		
		if ($sPathPartNew && in_array($sPathPartNew, $aBrotherPathPart))
			return ucfirst(ObjectTypeInfo::get($this->iNode)->sUnUne) . ' de mme parent utilise dj ce ' . $sFileOrDir;
		
		$sPathFullParent = path::concat(sDataPath, concatPathPart($aParent));
		$sPathFullOld = path::concat($sPathFullParent, $sPathPartOld);
		$sPathFullNew = path::concat($sPathFullParent, $sPathPartNew);
		
		$bOldExists = $sPathFullParent && $sPathPartOld && file_exists($sPathFullOld);
		$bNewExists = $sPathFullParent && $sPathPartNew && file_exists($sPathFullNew);
		
		if ($bNewExists) {
			// New path exists: no overwrite,
			// accept only if it is the previous path, unchanged
			
			if ($sPathPartNew != $sPathPartOld) {
				$sError = 'Le ' . $sFileOrDir . ' ' . html::kbdText($sPathFullOld) . ' existe dj&nbsp;: veuillez crer une entre dans DENIF ' . html::b('avant') . ' ';
				$sError .= ($bIsFile) ? 'duploader le fichier' : 'de crer le rpertoire';
				return $sError;
			}
			
		}else if ($sPathPartNew) {
			// New path specified: rename or create directory
			
			if ($bOldExists) {
				if (!@rename($sPathFullOld, $sPathFullNew))
					return 'Impossible de renommer le ' . $sFileOrDir . ' ' . html::kbdText($sPathFullOld) . ' en ' . html::kbdText($sPathFullNew);
				
			}else if (!$bIsFile && !@is_dir($sPathFullNew)) {
				if (!@path::mkdir($sPathFullNew, iDataUnixRightsDirs))
					return 'Impossible de crer le rpertoire ' . html::kbdText($sPathFullNew);
				@chgrp($sPathFullNew, sDataUnixGroup);
			}
			
		}else if ($bOldExists) {
			// New path empty: delete
			
			if (!(($bIsFile) ? @unlink($sPathFullOld) : @rmdir($sPathFullOld)))
				return 'Impossible de supprimer le ' . $sFileOrDir . ' ' . html::kbdText($sPathFullOld);
		}
		
		return null;
	}
	
	/// @brief Check than the given string is a valid value for #$sPathPart.
	/// It must not contain any special character, and not begin with a dot.
	public static function checkPathPart($sPathPart)
	{
		return !$sPathPart || ($sPathPart[0] != '.' &&
			!regexp::testChars(path::sPartInvalidCharsSpec, $sPathPart));
	}
	
	
	/// @brief Try to modify #$sPathPart according to the URL of the document.
	/// If #$sPathPart can be changed, call onPathPartChanged() to update the file system.
	/// #$sURL is cleared if not an absolute URL.
	/// Return \c true if success, \c false otherwise.
	/// Can be used on documents only.
	///
	/// @param $aParent list of all parents of the node, excluding the node itself.
	/// Used to compute the full path of the node.
	/// @param $aBrotherPathPart path part of all brothers of the node, excluding the node itself.
	/// Used to prevent two brothers from having the same path part.
	public function setPathPartFromURL(array $aParent, array $aBrotherPathPart)
	{
		// Extract the file name of the URL
		$sPathPartNew = basename(@parse_url($this->sURL, PHP_URL_PATH));
		
		if ($sPathPartNew) {
			// Make sure the file name is valid
			$sPathPartNew = regexp::replaceChars(sDataPathPartInvalidCharsSpec, '_', $sPathPartNew);
			if (!$sPathPartNew || $sPathPartNew[0] == '.')
				$sPathPartNew = '_' . $sPathPartNew;
		}
		
		// Find a unique name for the path part,
		// if necessary by appending a number to it
		$iUnique = 1;
		do{
			if ($iUnique == 1)
				$sPathPartNewFinal = $sPathPartNew;
			else{
				list($sBase, $sExt) = path::splitExtension($sPathPartNew);
				$sPathPartNewFinal = $sBase . '_' . $iUnique . $sExt;
			}
			$iUnique++;
		}while ($sPathPartNewFinal && in_array($sPathPartNewFinal, $aBrotherPathPart));
		
		// Change the path part
		if ($this->onPathPartChanged($sPathPartNewFinal, $aParent, $aBrotherPathPart) !== null)
			return false;
		
		$this->sPathPart = $sPathPartNewFinal;
		$this->sURL      = $this->getAbsoluteURL();
		return true;
	}
	
	/// @}
}

node::$table = ObjectTableInfo::get(tableNode);


//-----------------------------------------------------------------
/// Base class for properties
//-----------------------------------------------------------------

abstract class property extends ObjectBase
{
	public $bValidated;   ///< Property is finalized, or still editable by guests?
	
	
	public function getFormName() { return 'property'; }
	
	
	public function link(PageDenif $page, $sTitle = null, $sClassName = '')
	{
		if ($sTitle === null)
			$sTitle = html::text($this->getLinkTitle());
		return $page->linkTarget($this->getLinkTarget(), $sTitle, $sClassName);
	}
	
	
	public function getLinks($iNode)
	{
		return sql::selectFetch(getList('link;node'),
			array('link.*'), sql::where(array(
				'iProperty'  => $this->getType(),
				'idProperty' => $this->id,
				'!node.id'   => 'idNode',
				'node.iNode' => $iNode,
			)));
	}
	
	
	protected static function formatListItem(PageDenif $page, $sTitle, array $aProperty, $sLinkClassName ='')
	{
		$aLink = array();
		foreach ($aProperty as $property)
			$aLink[] = $property->linkFull($page, $sLinkClassName);
		if (!$aLink)
			return '';
		
		if ($sTitle)
			$sTitle .= ' ';
		return html::span($sTitle . concat($aLink, ', '), attr::className('propertybar'));
	}
	
	protected static function formatListNoSubType(PageDenif $page, $iProperty, array $aProperty)
	{
		$info = ObjectTypeInfo::get($iProperty);
		$sTitle = ucfirst((count($aProperty) >= 2) ? $info->sPlural : $info->sAlone);
		$sTitle = html::span($sTitle . '&nbsp;:', attr::className('title'));
		return self::formatListItem($page, $sTitle, $aProperty);
	}
	
	
	//-----------------------------------------------------------------
	/// @name Deleting
	/// @{
	//-----------------------------------------------------------------
	
	public function delete()
	{
		$info = $this->getInfo();
		if (!$this->canDelete())
			return ucfirst($info->sLeLa) . ' est utilise par un nud.';
		
		sql::delete($info->table->sName, sql::whereId($this->id));
		return null;
	}
	
	public function canDelete()
	{
		return 0 == sql::count('link', sql::where(array(
			'iProperty'  => $this->getType(),
			'idProperty' => $this->id,
		)));
	}
	
	public function dumpBeforeDelete()
	{
		return self::dumpField($this->getInfo()->sAlone, $this->getLinkTitle());
	}
	
	/// @}
}


//-----------------------------------------------------------------
/// Cursus property
//-----------------------------------------------------------------

class cursus extends property
{
	public $sShortName, $sFullName;
	public $sURL;
	
	public function getType() { return propertyCursus; }
	public function getLinkTitle() { return defaultValue($this->sShortName, '[sans nom]'); }
	public function getLinkTarget() { return new target('index_cursus.php', array('id' => $this->id), 'id'); }
	
	public function sanityCheck()
	{
		$a = array();
		if (!$this->sShortName)  $a[] = 'nom vide';
		if (!$this->sFullName)   $a[] = 'nom long vide';
		if (!$this->sURL)        $a[] = 'URL vide';
		return $a;
	}
	
	public static function formatList(PageDenif $page, array $aProperty)
	{
		return self::formatListNoSubType($page, propertyCursus, $aProperty);
	}
}


//-----------------------------------------------------------------
/// Theme property
//-----------------------------------------------------------------

class theme extends property
{
	public $sTitle;
	
	public function getType() { return propertyTheme; }
	public function getLinkTitle() { return defaultValue($this->sTitle, '[sans titre]'); }
	public function getLinkTarget() { return new target('index_themes.php', array('id' => $this->id), 'id'); }
	
	public function sanityCheck()
	{
		$a = array();
		if (!$this->sTitle)  $a[] = 'titre vide';
		return $a;
	}
	
	public static function formatList(PageDenif $page, array $aProperty)
	{
		return self::formatListNoSubType($page, propertyTheme, $aProperty);
	}
}


//-----------------------------------------------------------------
/// Human property: author, teacher, TD-man, contributor, ...
//-----------------------------------------------------------------

class human extends property
{
	public $sName;
	public $sURL;
	public $sEmail;
	
	public function getType() { return propertyHuman; }
	public function getLinkTitle() { return defaultValue($this->sName, '[sans nom]'); }
	public function getLinkTarget() { return new target('teacher.php', array('id' => $this->id)); }
	public function getLinkOutClassName() { return 'home'; }
	public function getLinkOutToolTip() { return 'Page perso'; }
	
	public function sanityCheck()
	{
		$a = array();
		if (!$this->sName)   $a[] = 'nom vide';
		if (!$this->sURL)    $a[] = 'URL page perso vide';
		if (!$this->sEmail)  $a[] = 'e-mail vide';
		return $a;
	}
	
	public static function formatList(PageDenif $page, array $aProperty)
	{
		return self::formatListItem($page, '', $aProperty, 'human');
	}
}


/// @brief Return the node associated to an ID,
/// \c null if the ID is invalid or if the node is temporary.
function getNode($id, $sMore = '')
{
	return ($id)
		? sql::selectOneFetch(node::$table->sName, array('*'), sql::whereId($id) .
			' AND iFinal<=' . finalNo . ' ' . $sMore, node::$table->sName)
		: null;
}

/// @brief Return all non-temporary nodes having the given type,
/// indexed by their ID and correctly sorted.
function getNodes($iNode, $sMore = '')
{
	$sTable = node::$table->sName;
	$sMore .= ' ORDER BY ' . ObjectTypeInfo::get($iNode)->sOrderBy;
	return sql::selectFetchBy($sTable, 'id', array('*'),
		sql::whereOne('iNode', $iNode) .
		' AND iFinal<=' . finalNo . ' ' . $sMore, $sTable);
}

/// Get the parent nodes of a nodes set.
function getParentNodes($iNodeParent, array $aNode, $sMore = '')
{
	if (!$aNode)
		return array();
	return getNodes($iNodeParent, 'AND ' .
		sql::formatInIntList('id', projectToUnique($aNode, 'idParent')));
}


/// Get all properties of the given type, indexed by type then by ID.
function getProperties($iProperty, $sMore = '')
{
	$info = ObjectTypeInfo::get($iProperty);
	$sTable = $info->table->sName;
	$sMore .= ' ORDER BY ' . $info->sOrderBy;
	return sql::selectFetchBy($sTable, 'id', array('*'), $sMore, $sTable);
}

function getPropertiesOf($iProperty, $sMore = '', array $aField = array('prop.*'), array $aTableMore = array())
{
	$info = ObjectTypeInfo::get($iProperty);
	$sTable = $info->table->sName;
	$sMore .= ' ORDER BY ' . $info->sOrderBy;
	$aField[] = 'idNode';
	$aTableMore[]       = 'link';
	$aTableMore['prop'] = $sTable;
	return sql::selectFetch($aTableMore,
		$aField, sql::where(array(
		'link.iProperty'   => $iProperty,
		'!link.idProperty' => 'prop.id',
	)) . ' ' . $sMore, $sTable);
}

function getPropertiesAndLinkOf($iProperty, $sMore = '', array $aTableMore = array())
{
	return getPropertiesOf($iProperty, $sMore,
		array('idNode', 'iProperty', 'idProperty', 'prop.*', 'idLink' => 'link.id'),
		$aTableMore);
}


/// @brief Get all properties in an array of properties array,
/// indexed by type then by ID.
function getAllProperties()
{
	$a = array();
	for ($iProperty = propertyFirst; $iProperty <= propertyLast; $iProperty++)
		$a[$iProperty] = getProperties($iProperty);
	return $a;
}


function node2property($iNode, $iProperty)
{
	return assocNodeProperty($iNode, $iProperty, $iProperty);
}

function property2node($iProperty, $iNode)
{
	return assocNodeProperty($iNode, $iProperty, $iNode);
}

function assocNodeProperty($iNode, $iProperty, $iValue)
{
	$bValueNode = isDefineInRange($iValue, 'node');
	$aTable = getList('link;node');
	$aWhere = array(
		'node.iNode' => $iNode,
		'iProperty'  => $iProperty,
		'!node.id'   => 'link.idNode',
	);
	
	if (!$bValueNode) {
		$aTable['prop'] = ObjectTypeInfo::get($iProperty)->table->sName;
		$aWhere['!prop.id'] = 'link.idProperty';
	}
	
	$aLink = sql::selectFetch($aTable,
		getList('idNode;idProperty'), sql::where($aWhere) .
		' ORDER BY ' . (($bValueNode) ? 'node' : 'prop') . '.' . ObjectTypeInfo::get($iValue)->sOrderBy);
	return assocFieldBy($aLink,
		($bValueNode) ? 'idProperty' : 'idNode',
		($bValueNode) ? 'idNode' : 'idProperty');
}


/// Node table alias used by #getNodeTable().
function getNodeTable($iNode)
{
	return 'n' . $iNode;
}

/// @brief Get the SQL tables and conditions necessary to browse the nodes tree.
///
/// Parent/child links of each node type between \p $iNodeRoot and \p $iNodeLeaf
/// are established. For a given leaf node, each <tt>getNodeTable($iNode)</tt> table
/// contains the data about the parent of type \c $iNode.
/// Nodes being created (\c stateCreating) are ignored.
function sqlNodesChildren(&$aTable, &$aWhere, $iNodeRoot, $iNodeLeaf)
{
	for ($iNode = $iNodeLeaf; $iNode >= $iNodeRoot; $iNode--) {
		$sName = getNodeTable($iNode);
		$aTable[$sName] = node::$table->sName;
		$aWhere[$sName . '.iNode'] = $iNode;
		$aWhere[$sName . '.iFinal <= ' . finalNo] = null;
		if ($iNode < $iNodeLeaf)
			$aWhere['!' . $sName . '.id'] = getNodeTable($iNode + 1) . '.idParent';
	}
}


/// @brief Return an array containing sorted, unique, elements
/// from a string <tt>'id1,id2,...,idN'</tt>.
function parseNodeIdList($sNodes)
{
	$a = array_unique(array_filter(array_map('intval', explode(',', $sNodes))));
	sort($a);
	return $a;
}


/// Modes proposed in the right pane of the Ajax documents interface.
/// - <tt>webModeAddDocs</tt>: when a link is clicked, display a form
///   asking for its title; by default, the documents cursor is at the end:
///   the documents will be added at the end.
/// - <tt>webModeNavigate</tt>: do not catch links clicks,
///   allowing to navigate in the right pane as in a normal page.
/// - <tt>webModeEditLinks</tt>: when a link is clicked, do not display
///   a form but set the URL of the current document,
///   then move the documents cursor to the next document.
defineEnum('webMode', 'AddDocs;Navigate;EditLinks', true);

define('webModeDefault', webModeAddDocs);


?>