<?php defined('KERNEL_INCLUDED') or die('kernel should be included first');
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder, Arthur Charguraud

//-----------------------------------------------------------------
/// @file
/// @brief #Maintenance: maintenance services for database and cached files
//-----------------------------------------------------------------


//-----------------------------------------------------------------
/// @brief Maintenance services for database and cached files
//-----------------------------------------------------------------

class Maintenance
{
	private $page;                        ///< Page the results are printed in, may be \c null.
	private $targetPage = null;           ///< Used as exit URL by edit links, \c null if $page is \c null.
	
	private $aNode = null;                ///< All nodes, stored there to avoid multiple loading.
	private $aNoPathPartByID = array();   ///< Nodes encountered so far having an invalid node::$sPathPart.
	
	
	/// @brief Perform minor repairs required by many other maintenance operations.
	public function __construct($page = null)
	{
		$this->page = $page;
		
		sql::update(node::$table->sName, array('idParent' => 0), sql::whereOne('iNode', nodeFirst));
	}
	
	
	/// @brief Accepted maintenance operations,
	/// in the order they should be processed if several are given,
	/// and indexed by a valid URL parameter name.
	public function getOperations()
	{
		return array(
			'bExport'      => 'databaseExport',
			'bClean'       => 'databaseClean',
			'bOptimize'    => 'databaseOptimize',
			'bOrder'       => 'repairOrder',
			'bFindInvalid' => 'findInvalid',
			'bCachedFiles' => 'checkCachedFiles',
			'bDirectories' => 'checkDirectories',
		);
	}
	
	
	/// @brief Execute the operation associated to the given parameter.
	/// Return the resulting messages, in HTML if $page is not \c null, in plain text otherwise.
	public function executeOperation($sParam)
	{
		$sFunction = defaultElement($this->getOperations(), $sParam, '');
		if (!$sFunction)
			return '';
		
		// Make $targetPage pointing to $page with \p $sParam set to \c true
		$page = $this->page;
		$this->targetPage = ($page)
			? $page->getSelfTarget(targetSession, array($sParam => true))
			: null;
		
		return $this->$sFunction();
	}
	
	
	/// @brief Compute database statistics.
	public function getStatistics()
	{
		$s = '';
		
		$aStat = array();
		$aNode = $this->loadNodes();
		foreach (getDefineRange('node') as $iNode) {
			$nbTotal   = 0;
			$nbVisible = 0;
			foreach ($aNode as $node) {
				if ($node->iNode == $iNode) {
					$nbTotal++;
					if ($node->aFlagInherit[nodeFlagVisible] == visibleYes)
						$nbVisible++;
				}
			}
			$aStat[] = array(ucfirst(ObjectTypeInfo::get($iNode)->sPlural), $nbVisible, $nbTotal, sprintf('%d %%', $nbVisible * 100 / $nbTotal));
		}
		
		$aStatProperty = array();
		foreach (getDefineRange('property') as $iType) {
			$info = ObjectTypeInfo::get($iType);
			$aStat[] = array(ucfirst($info->sPlural), '', sql::count($info->table->sName), '');
		}
		
		return html::table($aStat, html::colClass('stats-name') . str_repeat(html::colClass('stats-value'), 3), attr::className('stats'), html::trthList(array('Type', 'Visible', 'Total', 'Ratio')));
	}
	
	
	/// @brief Export the whole database as gzip'd SQL,
	/// by using 'mysqldump' and 'gzip'.
	public function databaseExport()
	{
		page::uploadData(sql_sBase . '.sql');
		sql::dumpTables();
		exit;
	}
	
	
	/// @brief Clean the database: delete old sessions, delete temporary nodes,
	/// trim text fields.
	public function databaseClean()
	{
		// Delete old sessions
		SessionManager::clean();
		
		// Delete temporary nodes
		sql::delete(node::$table->sName, sql::whereOne('iFinal', finalTemp));
		
		// Trim text fields
		$aTextField = array(
			'cursus' => 'sShortName;sFullName;sURL',
			'human'  => 'sName;sURL;sEmail',
			'theme'  => 'sTitle',
			'node'   => 'sTitle;sDetails;sPathPart;sURL',
		);
		foreach ($aTextField as $sTable => $sFields) {
			$aUpdate = array();
			foreach (getList($sFields) as $sField)
				$aUpdate['!' . $sField] = 'TRIM(' . $sField . ')';
			sql::update($sTable, $aUpdate);
		}
		
		return $this->success('Base de donnes nettoye.');
	}
	
	
	/// @brief Optimize the database: defragment all tables.
	public function databaseOptimize()
	{
		sql::optimizeTables();
		return $this->success('Base de donnes optimise.');
	}
	
	
	/// @brief Repair the order of the nodes:
	/// ensure that the N children of all parent nodes have order index from 0 to N-1.
	public function repairOrder()
	{
		sql::lockTables(array(node::$table->sName));
		
		$this->loadNodes();
		$aNodeByParent = assocBy($this->aNode, 'idParent');
		
		$aError = array();
		foreach ($aNodeByParent as $idParent => $aChild) {
			$nodeParent = defaultElement($this->aNode, $idParent, null);
			if ($nodeParent)
				$aError += $this->repairOrderOfChildren($nodeParent, $aChild);
		}
		
		sql::unlockTables();
		
		ksort($aError);
		return $this->success('Ordre des lments rpar.', concat($aError));
	}
	
	private function repairOrderOfChildren(node $nodeParent, array $aChild)
	{
		// Check that children of $nodeParent have order from 0 to N-1
		$aOrder = projectTo($aChild, 'iOrder');
		sort($aOrder);
		$bOK = (array_keys($aOrder) == array_values($aOrder));
		if ($bOK)
			return array();
		
		// Error: reorder the children
		// @I is a MySQL variable taking values from 0 to N-1
		sql::query('SET @I=-1');
		$aWhere = array('iNode' => $nodeParent->iNode + 1);
		if ($nodeParent->id)
			$aWhere['idParent'] = $nodeParent->id;
		sql::update(node::$table->sName,
			array('!iOrder' => '@I:=@I+1'), sql::where($aWhere) . ' ORDER BY iOrder');
		
		return array($nodeParent->sPathFull =>
			$this->messageObject($nodeParent, $nodeParent->sTitle, 'Ordre des fils rpar'));
	}
	
	
	public function findInvalid()
	{
		$s = '';
		
		$aNode = $this->loadNodes();
		
		$aLink = sql::selectFetch('link', array('*'));
		$aProperty = getAllProperties();
		
		// Check links, create associative arrays
		$aPropertyUsage = array();
		$aPropertyTypeUsage = array();
		foreach ($aLink as $link) {
			$aPropertyUsage    [$link->iProperty][$link->idProperty] =
			$aPropertyTypeUsage[$link->idNode]   [$link->iProperty]  = true;
			
			if (!isset($aNode[$link->idNode])) {
				$s .= $this->message('Lien attach  un nud inexistant',
					sprintf('ID lien = %d, ID = nud %d', $link->id, $link->idNode));
				sql::delete('link', sql::whereId($link->id));
			}
		}
		
		// Check nodes
		foreach ($aNode as $node) {
			$info = $node->getInfo();
			
			// Check that nodes are attached to at least one property
			// of each allowed kind, except human property which is ignored
			$aNodePropertyTypeUsage = defaultElement($aPropertyTypeUsage, $node->id, array());
			foreach ($info->aProperty as $iProperty => $nbAdd)
				if ($iProperty != propertyHuman && !isset($aNodePropertyTypeUsage[$iProperty]))
					$s .= $this->messageObject($node, $node->sTitle, 'Aucun(e) ' . ObjectTypeInfo::get($iProperty)->sAlone . ' associ(e) ');
		}
		
		// Find orphan properties
		foreach ($aProperty as $iProperty => $aPropertyCur) {
			$aCurPropertyUsage = defaultElement($aPropertyUsage, $iProperty, array());
			$sMessage = 'Proprit inutilise, ' . ObjectTypeInfo::get($iProperty)->sAlone;
			foreach ($aPropertyCur as $property)
				if (!isset($aCurPropertyUsage[$property->id]))
					$s .= $this->messageObject($property, $property->getLinkTitle(), $sMessage);
		}
		
		// Find duplicate teachers, i.e. couple of teachers having similar names
		// according to the Levenshtein distance
		$aHuman = defaultElement($aProperty, propertyHuman, array());
		$aHumanError = array();
		foreach ($aHuman as $human1) {
			foreach ($aHuman as $human2) {
				if ($human1->id < $human2->id &&
				    levenshtein($human1->sName, $human2->sName) <= levenshteinThresholdNames)
					$aHumanError[$human1->sName] =
						$this->messageObject($human1, $human1->getLinkTitle(), 'Enseignant probablement identique  celui ci-dessous') .
						$this->messageObject($human2, $human2->getLinkTitle(), 'Enseignant probablement identique  celui ci-dessus');
			}
		}
		ksort($aHumanError);
		$s .= concat($aHumanError);
		
		// Perform sanity check on all nodes and properties
		$aObject = array_merge($aNode, flatten($aProperty));
		$aObjectError = array();
		foreach ($aObject as $obj) {
			$aError = $obj->sanityCheck();
			if ($aError)
				$aObjectError[$obj->getType() . ' ' . $obj->getLinkTitle()] = $this->messageObject($obj, $obj->getLinkTitle(), 'Objet invalide, ' . concat(', ', $aError));
		}
		ksort($aObjectError);
		$s .= concat($aObjectError);
		
		return $this->success('lments invalides vrifis.', $s);
	}
	
	
	/// @brief Repair the order of the nodes:
	/// ensure that the N children of all parent nodes have order index from 1 to N.
	public function checkCachedFiles()
	{
		$aNotCached = array();
		foreach ($this->loadNodes() as $node) {
			if ($node->iNode != nodeLast)
				continue;
			
			$sPath = $this->getFullPath(false, $s, $node, $nodeParent);
			if ($sPath === false)
				continue;
			
			$bCached = is_file($sPath);
			if ($node->bCached != $bCached)
				sql::update(node::$table->sName, array('bCached' => $bCached),
					sql::whereId($node->id));
			
			if (!$bCached)
				$aError[$sPath] = $this->messageObject($node, $sPath, 'Fichier absent du cache');
		}
		
		ksort($aError);
		return $this->success('Fichiers du cache vrifis.', concat($aError));
	}
	
	
	/// @brief If necessary, create the directories pointed by node::$sPathPart exist
	/// and give them the expected Unix rights and group.
	public function checkDirectories()
	{
		$s = '';
		
		if (!bWindows)
			system('./set-data-rights.pl', $iExitCode);
		
		foreach ($this->loadNodes() as $node) {
			$sPath = $this->getFullPath(true, $s, $node, $nodeParent);
			if ($sPath === false)
				continue;
			
			if ($node->iNode == nodeLast)
				continue;
			
			if (is_dir($sPath)) {
				@chmod($sPath, iDataUnixRightsDirs);
				@chgrp($sPath, sDataUnixGroup);
				
			}else{
				$s .= $this->messageObject($nodeParent, $sPath, 'Rpertoire non-cr');
				if (file_exists($sPath))
					$s .= $this->messageObject($nodeParent, $sPath, 'Un fichier a le nom du rpertoire');
				else if (@path::mkdir($sPath, iDataUnixRightsDirs, true))
					@chgrp($sPath, sDataUnixGroup);
				else
					$s .= $this->messageObject($nodeParent, $sPath, 'Impossible de crer le rpertoire');
			}
		}
		
		return $this->success('Rpertoires vrifis.', $s);
	}
	
	
	/// @brief Load all nodes, if not done yet. Return the nodes array.
	private function loadNodes()
	{
		if (!$this->aNode) {
			$sTable = node::$table->sName;
			$aNode = array();
			$nodeDefault = node::getDefault();
			$nodeDefault->sPathFull = '';
			foreach (getDefineRange('node') as $iNode) {
				$aNodeOfType = sql::selectFetchBy($sTable, 'id', array('*'),
					sql::whereOne('iNode', $iNode), $sTable);
				foreach ($aNodeOfType as $node) {
					if ($iNode == nodeFirst)
						$nodeParent = $nodeDefault;
					else if (isset($aNode[$node->idParent]))
						$nodeParent = $aNode[$node->idParent];
					else
						continue;
					
					$node->computeInheritedFlags($nodeParent);
					$aNode[$node->id] = $node;
				}
			}
			
			$this->aNode = $aNode;
		}
		return $this->aNode;
	}
	
	
	/// @brief Return the full path of a node, by using the full path of its parent,
	/// or \c false if none is available.
	/// Append error messages to \p $s if \p $bVerbose is \c true.
	private function getFullPath($bVerbose, &$s, node $node, &$nodeParent)
	{
		if (!$node->sPathPart)
			return false;
		
		$aPart = array();
		$nodeParent = $node;
		$bOK = true;
		for (;;) {
			if (!$nodeParent->sPathPart) {
				if (!isset($this->aNoPathPartByID[$nodeParent->id])) {
					$this->aNoPathPartByID[$nodeParent->id] = true;
					if ($bVerbose)
						$s .= $this->messageObject($nodeParent, $nodeParent->getLinkTitle(),
							'Le parent na pas de rpertoire');
				}
				$bOK = false;
			}
			
			$aPart[] = $nodeParent->sPathPart;
			if (!$nodeParent->idParent)
				break;
			$nodeParent = $this->aNode[$nodeParent->idParent];
		}
		
		if (!$bOK)
			return false;
		
		$aPart[] = sDataPath;
		return path::concat(array_reverse($aPart));
	}
	
	
	/// @brief Return a message.
	private function message($sMessage, $sInfo)
	{
		$page = $this->page;
		return ($page)
			? html::div($sMessage . '&nbsp;: ' . $sInfo)
			: str_replace('', '\'', $sMessage . ' : ' . $sInfo . CRLF);
	}
	
	/// @brief Return a message about an object.
	private function messageObject(ObjectBase $obj, $sObjectInfo, $sMessage)
	{
		$page = $this->page;
		return $this->message($sMessage, ($page)
			? html::kbdText($sObjectInfo) . ' ' .
				$page->linkEdit(getLinkEdit($obj, clone $this->targetPage))
			: $sObjectInfo . ' (ID = ' . $obj->id . ')' . CRLF);
	}
	
	/// @brief Format an OK message, in a paragraph tag.
	/// Return nothing if the maintenance object is not attached to a page.
	private function success($sMessage, $sDetails = '')
	{
		if ($this->page) {
			if ($sDetails)
				$sDetails = html::par($sDetails);
			$sDetails .= html::par($sMessage, attr::className('ok'));
		}else
			$sDetails .= $sMessage . CRLF;
		
		return $sDetails;
	}
}

?>