<?php defined('KERNEL_INCLUDED') or die('kernel should be included first');
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder, Arthur Charguraud

//-----------------------------------------------------------------
/// @file
/// @brief #SessionManager: session state storing.
//-----------------------------------------------------------------


defineEnum('level', 'Guest;Moderator;Admin');


//-----------------------------------------------------------------
/// @brief Store the session state: user name, rights, edit mode
///
/// To add an admin:
/// SessionManager::registerUser('username', 'password', levelAdmin);
//-----------------------------------------------------------------

class SessionManager
{
	private $user, $session;   // Database objects
	private $bLogged;
	private $sLogoutMessage;
	private $bSimulateGuest = false;
	
	
	/// Process login, session loading, logout
	public function __construct()
	{
		$_REQUEST->prepare('bLogin;bLogout;sLogin;sPassword;bCookie;sSession;bSimulateGuest');
		
		$this->setLogoutMessage('Veuillez vous authentifier');
		
		if (!$this->loadSession($_REQUEST->sSession))
			$this->loadCookie($_COOKIE->getSafe('sCode'), false);
		
		if ($_REQUEST->bLogin)
			$this->login($_REQUEST->sLogin, $_REQUEST->sPassword, $_REQUEST->bCookie);
		else if ($_REQUEST->bLogout)
			$this->logout();
		
		$this->bSimulateGuest = $_REQUEST->bSimulateGuest;
		
		// Editing mode: refresh at every click
		if ($this->isEditing())
			header('Cache-Control: no-cache, must-revalidate');
	}
	
	
	//-----------------------------------------------------------------
	/// @name High-level actions
	/// @{
	//-----------------------------------------------------------------
	
	/// Create a new session from a login and password.
	private function login($sLogin, $sPassword, $bCookie)
	{
		$this->user = sql::selectOneFetch('user_account', array('*'), sql::where(array(
			'sLogin'       => $sLogin,
			'sPasswordMD5' => md5($sPassword),
		)));
		$this->bLogged = (bool)$this->user;
		if (!$this->bLogged) {
			$this->setLogoutMessage('Erreur dauthentification');
			return;
		}
		
		$this->createSession($_REQUEST->getSafe('bSetEdit'));
		
		if ($bCookie)
			$this->createCookie();
		else
			$this->clearCookie();
	}
	
	
	/// Delete the current session.
	public function logout()
	{
		if ($this->bLogged) {
			sql::delete('user_session', sql::whereOne('sCode', $this->session->sCode));
			$this->setLogoutMessage('Session termine, merci de votre visite.');
		}
		
		$this->clearCookie();
	}
	
	
	/// Continue an existing session, return \c false if error.
	private function loadSession($sCode)
	{
		if (!$sCode)
			return false;
		
		$sWhere = sql::whereOne('sCode', $sCode);
		
		// Get the session from its code
		$session = $this->session = sql::selectOneFetch('user_session', array('*'), $sWhere);
		if (!$session) {
			$this->setLogoutMessage('Lidentifiant de session est invalide');
			return false;
		}
		
		// Test for expiration
		$dtNow = time();
		if (($dtNow - $session->dtLast) > timeoutSessionSeconds) {
			sql::delete('user_session', $sWhere);
			$this->setLogoutMessage('La session a expir');
			return false;
		}
		
		// Update last query time and edit flag
		$aSession = array('dtLast' => $dtNow);
		if (isset($_REQUEST->bSetEdit))
			$aSession['bEdit'] = $_REQUEST->getSafe('bSetEdit');
		foreach ($aSession as $sField => $value)
			$session->$sField = $value;
		sql::update('user_session', $aSession, $sWhere);
		
		// Load the user
		return $this->loadUser($session->idUser);
	}
	
	
	/// Start a new session from a cookie.
	private function loadCookie($sCode, $bEdit)
	{
		if (!$sCode)
			return;
		$sWhere = sql::whereOne('sCode', $sCode);
		
		// Get the cookie from its code
		$cookie = sql::selectOneFetch('user_cookie', array('*'), $sWhere);
		if (!$cookie)
			return;
		
		// Test for expiration, delete the cookie (will be created again later)
		$dtNow = time();
		$bExpired = ($dtNow - $cookie->dtLast) > timeoutCookieSeconds;
		sql::delete('user_cookie', $sWhere);
		if ($bExpired)
			return;
		
		// Create the session
		if (!$this->loadUser($cookie->idUser))
			return false;
		$this->createSession($bEdit);
		$this->createCookie();
		
		return true;
	}
	
	
	public static function registerUser($sLogin, $sPassword, $iLevel = levelGuest, $iWebMode = webModeDefault)
	{
		sql::insertInto('user_account', array(
			'sLogin'       => $sLogin,
			'sPasswordMD5' => md5($sPassword),
			'iLevel'       => $iLevel,
			'iWebMode'     => $iWebMode));
	}
	
	
	/// @brief Switch to editing mode. Return \c false if impossible (need to be logged).
	public function wantEditing()
	{
		if ($this->isEditing())
			return true;
		if (!$this->bLogged)
			return false;
		
		$this->session->bEdit = true;
		sql::update('user_session', array('bEdit' => true), sql::whereOne('sCode', $this->session->sCode));
		return true;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Low-level actions
	/// @{
	//-----------------------------------------------------------------
	
	/// Logout and set the error message.
	private function setLogoutMessage($sLogoutMessage)
	{
		$this->sLogoutMessage = $sLogoutMessage;
		
		$this->session = null;
		$this->bLogged = false;
		$this->user    = null;
	}
	
	
	/// Create a new session for the current user, supposed loaded: <tt>$this->user != null</tt>.
	private function createSession($bEdit)
	{
		$session         = new stdClass;
		$session->idUser = $this->user->id;
		$session->sCode  = md5(uniqid(rand(), true));
		$session->dtLast = time();
		$session->bEdit  = $bEdit;
		sql::replaceObj('user_session', $session);
		
		$this->session = $session;
	}
	
	
	/// Create a new cookie entry for the current session,
	/// and setup the cookie accordingly.
	private function createCookie()
	{
		$aCookie = array(
			'sCode'  => md5(uniqid(rand(), true)),
			'dtLast' => time(),
			'idUser' => $this->user->id,
		);
		
		sql::replaceObj('user_cookie', $aCookie);
		
		$_COOKIE->sCode = $aCookie['sCode'];
	}
	
	
	/// Load an user, put it in $user and update $bLogged.
	private function loadUser($idUser)
	{
		$this->user = sql::selectOneFetch('user_account', array('*'), sql::whereId($idUser));
		$this->bLogged = (bool)$this->user;
		return $this->bLogged;
	}
	
	
	/// Delete the auto-login cookie.
	private function clearCookie()
	{
		unset($_COOKIE->sCode);
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Accessors
	/// @{
	//-----------------------------------------------------------------
	
	public function getLogoutMessage() { return $this->sLogoutMessage; }
	public function isLogged()    { return $this->bLogged; }
	public function getLogin()    { return ($this->bLogged) ? $this->user->sLogin : null; }
	public function isModerator() { return $this->bLogged && !$this->bSimulateGuest && $this->user->iLevel >= levelModerator; }
	public function isAdmin()     { return $this->bLogged && !$this->bSimulateGuest && $this->user->iLevel >= levelAdmin; }
	public function isEditing()   { return $this->bLogged && $this->session->bEdit; }
	public function getCode()     { return ($this->bLogged) ? $this->session->sCode : ''; }
	public function getWebMode()  { return ($this->bLogged) ? $this->user->iWebMode : webModeDefault; }
	
	
	public function targetAddSessionParams(target $target)
	{
		if ($this->bLogged) {
			$target->addParam('sSession', $this->session->sCode);
			$target->addParam('bEdit', $this->isEditing());
		}
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Permissions
	/// @{
	//-----------------------------------------------------------------
	
	public function hasRightConfirm()
	{
		return $this->isAdmin();
	}
	
	public function hasRightCreate($iObject, $nodeParent = null)
	{
		if (!$this->isLogged())
			return false;
		
		if (isDefineInRange($iObject, 'node')) {
			if ($nodeParent && $nodeParent->iNode != $iObject - 1)
				return false;
			
		}else if (isDefineInRange($iObject, 'property')) {
			
		}else
			return false;
		
		return true;
	}
	
	public function hasRightModify(ObjectBase $obj)
	{
		if (!$this->isLogged())
			return false;
		
		if ($obj instanceof node) {
			$iFinal = $obj->aFlagInherit[nodeFlagFinal];
			if (($iFinal == finalTemp) ||
			    ($iFinal == finalNo && $obj->aFlagInherit[nodeFlagVisible] == visibleYes))
				return true;
		}else if ($obj instanceof human && !$obj->bValidated)
			return true;
		
		return $this->isAdmin();
	}
	
	public function hasRightDelete(ObjectBase $obj)
	{
		return $this->hasRightModify($obj);
	}
	
	public function hasRightDeleteCreating(ObjectBase $obj)
	{
		return $this->isLogged();
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Sanity
	/// @{
	//-----------------------------------------------------------------
	
	public static function clean()
	{
		$dt = time();
		sql::delete('user_session', 'WHERE dtLast < ' . ($dt - timeoutSessionSeconds));
		sql::delete('user_cookie',  'WHERE dtLast < ' . ($dt - timeoutCookieSeconds));
	}
	
	/// @}
}

?>