<?php defined('KERNEL_INCLUDED') or die('kernel should be included first');

// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder, Arthur Charguraud

//-----------------------------------------------------------------
/// @file
/// @brief #view: nodes tree dumper
//-----------------------------------------------------------------


/// @brief Flags for view::dump() function.
///
/// - <tt>NotFinalOnly</tt>: display only non-final nodes (proposed nodes)
/// - <tt>SeparateDisplay</tt>: use a two-panel display, with nodes list
///   on the top panel and nodes data on the bottom panel, initially hidden.
///   Clicking a node in the list displays its data in the bottom panel.
/// - <tt>UseHighlighting</tt>: highlight the nodes specified by node::$aOnlyByID.
///   Used to highlight search results while displaying parents and children of those results.
/// - <tt>ForceChildren</tt>: force displaying all children of node::$aOnlyByID nodes,
///   recursively, even if they are not in node::$aOnlyByID.
defineFlags('view', 'NotFinalOnly;SeparateDisplay;UseHighlighting;ForceChildren');


//-----------------------------------------------------------------
/// @brief Nodes tree dumper: takes a nodes list
/// and dump it as an HTML tree, with the property of each node.
///
/// If the current user is not an admin, do not dump nodes having
/// limited display rights.
/// Allows to display only some nodes among the the loaded nodes.
//-----------------------------------------------------------------

class view
{
	private $iMinNodeCanOrder;
	private $page;
	private $aNode, $aProperty, $aLink;
	private $bLoaded = false;
	
	private $bAdmin, $bEdit;
	private $bFilter;
	private $iFlags;
	private $aOnlyByID = null;
	
	private $nbFound;
	
	// Stack-related values
	private $aParent = array();      // iNode => ancestor of kind $iNode
	private $aLastOrder = array();   // iNode => last order index for elements of kind $iNode
	
	
	public function __construct(PageDenif $page)
	{
		$this->page = $page;
	}
	
	
	public function load($iNodeRoot, array $aNode, $aProperty = null, $aLink = null)
	{
		$this->iMinNodeCanOrder = $iNodeRoot + 1;
		
		if ($aProperty === null)
			$aProperty = getAllProperties();
		
		if ($aLink === null)
			$aLink = sql::selectFetch('link');
		
		$this->aNode     = $aNode;
		$this->aProperty = $aProperty;
		$this->aLink     = $aLink;
		$this->bLoaded   = true;
		
		$this->nbFound   = 0;
	}
	
	
	public function setFoundNodes(array $aOnlyByID)
	{
		$this->aOnlyByID = $aOnlyByID;
	}
	
	public function getFoundCount()
	{
		return $this->nbFound;
	}
	
	
	public function loadNodes($aOnlyByID = null)
	{
		$aNode = array();
		foreach (getDefineRange('node') as $iNode)
			$aNode[$iNode] = getNodes($iNode);
		
		$this->aOnlyByID = $aOnlyByID;
		$this->load(nodeFirst - 1, $aNode);
	}
	
	
	public function loadRessources($aidResource = null)
	{
		$aNode = array();
		foreach (getDefineRange('node') as $iNode)
			$aNode[$iNode] = getNodes($iNode, ($iNode == nodeFirst)
				? ' AND ' . sql::formatInIntList('id', $aidResource) : '');
		
		$this->load(nodeFirst - 1, $aNode);
	}
	
	
	public function loadNode(node $node)
	{
		$aNode = array_fill_range(nodeFirst, nodeLast, null);
		$aLink = $node->getLinksRecurse();
		
		// Parent nodes and the node itself
		$node->loadAllParents(false);
		for ($nodeParent = $node; $nodeParent; $nodeParent = $nodeParent->parent) {
			$aNode[$nodeParent->iNode] = array($nodeParent);
			$aLink += sql::selectFetchBy('link', 'id', array('*'),
				sql::whereOne('idNode', $nodeParent->id));
		}
		
		// Children nodes
		for ($iNode = $node->iNode + 1; $iNode <= nodeLast; $iNode++)
			$aNode[$iNode] = $node->loadChildren($iNode);
		
		$this->load($node->iNode, $aNode, null, $aLink);
	}
	
	
	public function getDisplayedResources()
	{
		$aResourceDisplayed = array();
		foreach ($this->aNode[nodeResource] as $resource)
			if ($resource->bDisplayed)
				$aResourceDisplayed[] = $resource;
		return $aResourceDisplayed;
	}
	
	
	private function isNodeNeeded(node $node)
	{
		if (!$this->bFilter)
			return true;
		
		if (($this->iFlags & viewNotFinalOnly) && $node->aFlagInherit[nodeFlagFinal] != finalNo)
			return false;
		
		if ($this->aOnlyByID == null || 
		    array_key_exists($node->id, $this->aOnlyByID)) {
			$this->nbFound++;
			return true;
		}
		
		return ($this->iFlags & viewForceChildren) &&
			$node->iNode > nodeFirst && $this->aParent[$node->iNode - 1]->bNeeded;
	}
	
	
	public function dump($iFlags = 0, $iTitleLevel = 2)
	{
		if (!$this->bLoaded)
			debug::fatalError('view must be loaded: see view::load() method');
		
		$this->bAdmin  = $this->page->isAdmin();
		$this->bEdit   = $this->page->isEditing();
		$this->iFlags  = $iFlags;
		$this->bFilter = ($iFlags & viewNotFinalOnly) || ($this->aOnlyByID !== null);
		
		$aResource = $this->aNode[nodeResource];
		if (!$aResource)
			return '';
		
		$this->aLastOrder[nodeResource] = end($aResource)->iOrder;
		
		// Dump the nodes tree, starting from the resources
		$nodeDefault = node::getDefault();
		$s = '';
		$bCanZip = false;
		foreach ($aResource as $resource) {
			$resource->parent = $nodeDefault;
			$s .= $this->dumpNode($iTitleLevel, $resource, $bCanZip);
		}
		return html::div($s, attr::className('nodes'));
	}
	
	
	private function dumpNode($iTitleLevel, node $node, &$bCanZip)
	{
		if ($node->iFinal == finalTemp)
			return '';
		
		$page = $this->page;
		$iNode = $node->iNode;
		$this->aParent[$iNode] = $node;
		
		$node->computeInheritedFlags($node->parent);
		if (!$this->bAdmin && $node->aFlagInherit[nodeFlagVisible] != visibleYes)
			return null;
		
		if ($iNode == nodeDocument)
			return $this->dumpDocument($node, $bCanZip);
		
		$iNodeChild = $iNode + 1;
		$infoChild = ObjectTypeInfo::get($iNodeChild);
		
		// Get the order index of the last child
		$aChild = array();
		$iOrderLast = 0;
		foreach ($this->aNode[$iNodeChild] as $nodeChild)
			if ($nodeChild->idParent == $node->id) {
				$aChild[] = $nodeChild;
				$iOrderLast = max($iOrderLast, $nodeChild->iOrder);
			}
		$this->aLastOrder[$iNodeChild] = $iOrderLast;
		
		$bAddInsertLinks = (!$this->bFilter && $iNodeChild >= $this->iMinNodeCanOrder &&
			$this->bEdit && $iNodeChild < nodeDocument &&
			$page->sessionManager->hasRightCreate($iNodeChild, $node));
		$target = $page->getSelfTarget(targetParams, array(), $node->getAnchor());
		
		$bNeeded = $node->bNeeded = $this->isNodeNeeded($node);
		
		$aChildText = array();
		$bCanZipChildren = false;
		if (!($this->iFlags & viewNotFinalOnly) || ($this->iFlags & viewForceChildren)) {
			$iOrder = 0;
			foreach ($aChild as $nodeChild)
				if ($nodeChild->idParent == $node->id) {
					$nodeChild->parent = $node;
					$sChild = $this->dumpNode($iTitleLevel + 1, $nodeChild, $bCanZipChildren);
					if (!$sChild)
						continue;
					
					if ($bAddInsertLinks)
						$aChildText[] = $this->separatorLinks($iNodeChild, $node, $iOrder++, $target);
					$aChildText[] = $sChild;
				}
		}
		$bCanZip |= $bCanZipChildren;
		
		$node->bDisplayed = $aChildText || $bNeeded;
		if (!$node->bDisplayed)
			return '';
		
		if ($bAddInsertLinks)
			$aChildText[] = $this->separatorLinks($iNodeChild, $node, -1, $target);
		
		list($sHeader, $sContents) = $this->dumpNodeHeader($iTitleLevel, $node,
			$bNeeded, $bCanZipChildren);
		
		switch ($iNode) {
			case nodeDocGroup:
				$sContents .= html::listU($aChildText);
				break;
			
			default:
				$sContents .= html::div(concat($aChildText), attr::className('nodes' . $node->iNode));
				break;
		}
		
		$s = $sHeader . $sContents;
		if (($this->iFlags & viewSeparateDisplay) && $iNode == nodeFirst)
			$s = html::div($s, attr::idClassName('f' . $node->id, 'nodeSD'));
		return $s;
	}
	
	private function separatorLinks($iNodeChild, node $node, $iOrder, target $target)
	{
		return html::div(
			$this->page->linkEdit(getLinkCreateChild($iNodeChild, $node->id, $iOrder++, $target)),
			attr::className('insert'));
	}
	
	
	/// Return array(0 => header, 1 => properties)
	private function dumpNodeHeader($iTitleLevel, node $node, $bNeeded, $bCanZip)
	{
		$page = $this->page;
		
		$aPropertyByType = array();
		foreach ($node->getInfo()->aProperty as $iProperty => $nbAdd)
			$aPropertyByType[$iProperty] = $this->getAssociatedProperties($iProperty, $node->id);
		
		$sHeader = $this->formatTitle($iTitleLevel, $node, $aPropertyByType, $bNeeded, $bCanZip);
		
		if ($node->sDetails)
			$sHeader .= html::div($node->formatDetails(), attr::className('node-details'));
		
		$aPropertyList = array();
		foreach ($aPropertyByType as $iProperty => $aProperty)
			if ($aProperty)
				$aPropertyList[] = call_user_func(array(ObjectTypeInfo::get($iProperty)->sName, 'formatList'),
					$page, $aProperty);
		
		return array($sHeader, concat('  ', $aPropertyList));
	}
	
	
	private function dumpDocument(node $node, &$bCanZip)
	{
		$node->bNeeded = $node->bDisplayed = $this->isNodeNeeded($node);
		
		if (!$node->bDisplayed)
			return null;
		
		$page = $this->page;
		
		$aLine = array();
		
		$aLine[] = concat($page->getNodeFlagsIcon($node));
		
		$sURL = $node->getAbsoluteURL();
		$sToolTip = '';
		$bCacheIcon = false;
		if ($node->sPathPart && ($this->bAdmin || $node->canGuestAccessCache())) {
			$dt = (is_file($node->sPathFull)) ? @filemtime($node->sPathFull) : 0;
			if ($dt) {
				$bCanZip = true;
				if (!$sURL) {
					$sToolTip = 'Seul le fichier en cache est disponible';
					if (!$this->bAdmin)
						$sURL = $node->sPathFull;
				}
				$aLine[] = $page->linkOut($node->sPathFull, html::nbsp(), 'action download', 'Mis en cache le ' . date('d/m/Y', $dt));
				$bCacheIcon = true;
			}
		}
		if (!$bCacheIcon)
			$aLine[] = html::span(html::nbsp(), attr::title('Absent du cache') + attr::className('action download grayed'));
		
		if ($this->bAdmin) {
			if ($sToolTip)
				$sToolTip .= CRLF;
			$sToolTip .= ltrim(html::text($node->sPathPart) . ' (ID = ' . $node->id . ')');
		}
		$aLine[] = $node->linkOutTo($page, $sURL, false, $sToolTip);
		
		return concat($aLine, ' ');
	}
	
	
	private function getAssociatedProperties($iProperty, $idNode)
	{
		$aProperty = array();
		foreach ($this->aLink as $link)
			if ($link->iProperty == $iProperty && $link->idNode == $idNode)
				$aProperty[] = $this->aProperty[$iProperty][$link->idProperty];
		return $aProperty;
	}
	
	
	private function formatTitle($iTitleLevel, node $node, array $aPropertyByType, $bNeeded, $bCanZip)
	{
		$page = $this->page;
		
		$info    = $node->getInfo();
		$sAnchor = $node->getAnchor();
		$id      = $node->id;
		$iNode   = $node->iNode;
		
		$nodeParent = ($iNode > nodeFirst) ? $node->parent : null;
		$iOrderLast = $this->aLastOrder[$iNode];
		
		$iFlags = objectLinksDetails | objectLinksDelete;
		if (!$this->bFilter && $iNode >= $this->iMinNodeCanOrder)
			$iFlags |= objectLinksMove;
		if ($bCanZip)
			$iFlags |= objectLinksZip;
		
		$targetDelete = $page->getSelfTarget(targetParams, array(), ($nodeParent) ? $nodeParent->getAnchor() : '');
		if ($targetDelete->getParam('id', 0) == $node->id)
			$targetDelete->addParam('id', ($nodeParent) ? $nodeParent->id : 0);
		$aLink = $page->getObjectLinks($node, $sAnchor, $iFlags, $iOrderLast, $targetDelete);
		append($aLink, $page->getNodeFlagsIcon($node));
		
		$aAttr = attr::id($sAnchor);
		if ($bNeeded && $this->bFilter && ($this->iFlags & viewUseHighlighting))
			$aAttr += attr::className('needed');
		
		$sTitleDetails = '';
		if ($this->bAdmin && $this->bEdit && $node->sPathPart)
			$sTitleDetails .= ' ' . html::kbdText($node->sPathPart, attr::className('pathpart'));
		
		return html::title($iTitleLevel,
			concat($aLink) . $node->linkFull($page) . $sTitleDetails, $aAttr);
	}
}

?>